import logging

from vja.list_service import ListService
from vja.model import Task, Label
from vja.urgency import Urgency

logger = logging.getLogger(__name__)


class TaskService:
    def __init__(self, list_service: ListService, urgency: Urgency):
        self._list_service = list_service
        self._urgency = urgency

    def task_from_json(self, task_json: dict) -> Task:
        list_object = self._list_service.find_list_by_id(task_json['list_id'])
        labels = Label.from_json_array(task_json['labels'])
        task = Task.from_json(task_json, list_object, labels)
        task.urgency = self._urgency.compute_for(task)
        return task



import functools
import re
import time
from datetime import datetime, timedelta

import dateutil.parser
from dateutil import tz
from parsedatetime import parsedatetime

_timedelta_regex = re.compile(r'^((?P<weeks>[.\d]+?)w)? *'
                              r'^((?P<days>[.\d]+?)d)? *'
                              r'((?P<hours>[.\d]+?)h)? *'
                              r'((?P<minutes>[.\d]+?)m)? *'
                              r'((?P<seconds>[.\d]+?)s?)?$')


def parse_date_arg_to_datetime(text: str):
    if not text:
        return None
    try:
        return dateutil.parser.isoparse(text)
    except ValueError:
        timetuple, pdt_context = parsedatetime.Calendar(version=parsedatetime.VERSION_CONTEXT_STYLE).parse(text)
        datetime_date = datetime.fromtimestamp(time.mktime(timetuple))
        if not pdt_context.hasTime:
            datetime_date = datetime_date.replace(hour=0, minute=0, second=0)
        return datetime_date


def parse_date_arg_to_iso(text: str):
    result = parse_date_arg_to_datetime(text)
    return result.astimezone(tz.tzlocal()).isoformat() if result else None


def parse_date_arg_to_timedelta(time_str: str):
    """
    Parse a time string e.g. '2h 13m' or '1.5d' into a timedelta object.
    Based on Peter's answer at https://stackoverflow.com/a/51916936/2445204
    and virhilo's answer at https://stackoverflow.com/a/4628148/851699
    :param time_str: A string identifying a duration, e.g. '2h13.5m'
    :return datetime.timedelta: A datetime.timedelta object
    """
    if not time_str:
        return None
    parts = _timedelta_regex.match(time_str)
    assert parts is not None, f"""Could not parse any time information from '{time_str}'.
    Examples of valid strings: '8h', '2d 8h 5m 2s', '2m4.3s'"""
    time_params = {name: float(param)
                   for name, param in parts.groupdict().items() if param}
    return timedelta(**time_params)


def parse_bool_arg(text: str):
    return text.lower() in ['true', '1', 't', 'y', 'yes'] if text else False


def rgetattr(obj, path: str, *default):
    attrs = path.split('.')
    try:
        return functools.reduce(getattr, attrs, obj)
    except AttributeError:
        if default:
            return default[0]
        raise


def parse_json_date(json_date):
    if json_date and json_date > '0001-01-02T00:00:00Z':
        return dateutil.parser.isoparse(json_date).astimezone(tz.tzlocal()).replace(tzinfo=None)
    return None



import dataclasses
import typing
from dataclasses import dataclass, field
from datetime import datetime, timedelta

from vja.parse import parse_json_date


def custom_output(cls):
    def str_function(self):
        """Returns a string containing only the non-null attribute values, excluding json attribute ."""
        return '\n'.join(f'{attribute.name}: {_str_value(getattr(self, attribute.name))}'
                         for attribute in dataclasses.fields(self)
                         if attribute.name != 'json' and getattr(self, attribute.name))

    def _str_value(attribute_value):
        return [_str_value(x) for x in attribute_value] if isinstance(attribute_value, list) else str(attribute_value)

    setattr(cls, '__str__', str_function)
    return cls


def data_dict(cls):
    def data_dict_function(self):
        return {k: _transform_value(v) for k, v in self.__dict__.items() if k != 'json'}

    def _transform_value(v):
        if _is_data_dict(v):
            return v.data_dict()
        if isinstance(v, list):
            return [_transform_value(x) for x in v]
        return v

    def _is_data_dict(v):
        return hasattr(v, 'data_dict') and callable(v.data_dict)

    setattr(cls, 'data_dict', data_dict_function)
    return cls


@dataclass(frozen=True)
@data_dict
class User:
    json: dict = field(repr=False)
    id: int
    username: str
    name: str
    default_list_id: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['username'], json['name'], json['settings']['default_list_id'])


@dataclass(frozen=True)
@data_dict
class Namespace:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    is_archived: bool

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'], json['description'], json['is_archived'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Namespace.from_json(x) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class List:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    is_favorite: bool
    is_archived: bool
    namespace: Namespace

    @classmethod
    def from_json(cls, json, namespace):
        return cls(json, json['id'], json['title'], json['description'], json['is_archived'], json['is_favorite'],
                   namespace)

    @classmethod
    def from_json_array(cls, json_array, namespace):
        return [List.from_json(x, namespace) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class Bucket:
    json: dict = field(repr=False)
    id: int
    title: str
    is_done_bucket: bool
    limit: int
    position: int
    count_tasks: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'],
                   json['is_done_bucket'],
                   json['limit'],
                   json['position'],
                   len(json['tasks']) if json['tasks'] else 0)

    @classmethod
    def from_json_array(cls, json_array):
        return [Bucket.from_json(x) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class Label:
    json: dict = field(repr=False)
    id: int
    title: str

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Label.from_json(x) for x in json_array or []]


@dataclass
@custom_output
@data_dict
# pylint: disable=too-many-instance-attributes
class Task:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    priority: int
    is_favorite: bool
    due_date: datetime
    reminder_dates: typing.List[datetime]
    repeat_mode: int
    repeat_after: timedelta
    start_date: datetime
    end_date: datetime
    percent_done: float
    done: bool
    done_at: datetime
    labels: typing.List[Label]
    tasklist: List
    position: int
    bucket_id: int
    kanban_position: int
    created: datetime
    updated: datetime
    urgency: float = field(init=False)

    @property
    def label_titles(self):
        return ",".join(map(lambda label: label.title, self.labels or []))

    @classmethod
    def from_json(cls, json, list_object, labels):
        return cls(json, json['id'], json['title'], json['description'],
                   json['priority'],
                   json['is_favorite'],
                   parse_json_date(json['due_date']),
                   [parse_json_date(reminder) for reminder in json['reminder_dates'] or []],
                   json['repeat_mode'],
                   timedelta(seconds=json['repeat_after']),
                   parse_json_date(json['start_date']),
                   parse_json_date(json['end_date']),
                   json['percent_done'],
                   json['done'],
                   parse_json_date(json['done_at']),
                   labels,
                   list_object,
                   json['position'],
                   json['bucket_id'],
                   json['kanban_position'],
                   parse_json_date(json['created']),
                   parse_json_date(json['updated'])
                   )

    def has_label(self, label):
        return any(x.id == label.id for x in self.labels)
